#include "rdma.h"
#include <rdma/rdma_verbs.h>
#include <fcntl.h>
#include <poll.h>
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#define err(msg, code)                                                         \
	fprintf(stderr, "%s:%d => msg: %s, code: %d, errno: %d, errmsg: %s\n", \
		__func__, __LINE__, msg, code, errno, strerror(errno))

struct mr_queue {
	struct ibv_mr **mr;
	size_t cap;
	size_t size;
};

static int __mr_cmp(const void *l, const void *r)
{
	const struct ibv_mr *lhs = *(const struct ibv_mr **)l;
	const struct ibv_mr *rhs = *(const struct ibv_mr **)r;

	if (lhs->addr > rhs->addr)
		return 1;

	return lhs->addr == rhs->addr ? 0 : -1;
}

static struct mr_queue *__queue_ctor(int n)
{
	if (n <= 0)
		n = 1;

	struct mr_queue *res =
		(struct mr_queue *)calloc(1, sizeof(struct mr_queue));
	res->size = 0;
	res->cap = n;
	res->mr = (struct ibv_mr **)calloc(res->cap, sizeof(struct ibv_mr *));
	return res;
}

static void __queue_append(struct mr_queue *q, struct ibv_mr *mr)
{
	if (q->size == q->cap) {
		q->cap *= 2;
		struct ibv_mr **mr = (struct ibv_mr **)calloc(
			q->cap, sizeof(struct ibv_mr *));
		for (size_t i = 0; i < q->size; ++i)
			mr[i] = q->mr[i];

		free(q->mr);
		q->mr = mr;
	}
	q->mr[q->size++] = mr;
}

static struct ibv_mr *__queue_find(struct mr_queue *q, void *buf, int len,
				   int *index)
{
	char *beg = (char *)buf;
	char *end = (char *)buf + len;
	int l = 0;
	int r = q->size - 1;

	while (l <= r) {
		int m = l + (r - l) / 2;
		char *mbeg = (char *)q->mr[m]->addr;
		char *mend = (char *)q->mr[m]->addr + q->mr[l]->length;

		if (mbeg > beg) {
			r = m - 1;
		} else if (mbeg < beg) {
			if (end < mend) {
				if (index)
					*index = m;
				return q->mr[m];
			}

			l = m + 1;
		} else {
			if (mbeg <= beg && mend >= end) {
				if (index)
					*index = m;
				return q->mr[m];
			}

			l += 1;
			r -= 1;
		}
	}
	return NULL;
}

static int __poll_ev(int fd, int count, int timeout)
{
	struct pollfd pfd = { .fd = fd, .events = POLLIN, .revents = 0 };
	int res = 0;

	do {
		res = poll(&pfd, 1, timeout);
		if (res == 0) {
			count -= 1;
			continue;
		}
		break;
	} while (count > 0);
	return res;
}

int ha_regmrv(struct ha_connector *ctx, struct iovec *iov, int n)
{
	struct mr_queue **q = NULL;
	struct ibv_pd *pd = NULL;

	if (ctx->acceptor) {
		struct ha_acceptor *a = ctx->acceptor;

		q = &a->mr_q;
		pd = a->cm_id->pd;
	} else {
		struct ha_connector *c = (struct ha_connector *)ctx;

		q = &c->mr_q;
		pd = c->cm_id->pd;
	}
	if (!*q)
		*q = __queue_ctor(n);

	for (int i = 0; i < n; ++i) {
		struct ibv_mr *mr =
			ibv_reg_mr(pd, iov[i].iov_base, iov[i].iov_len,
				   IBV_ACCESS_LOCAL_WRITE);
		if (!mr) {
			err("ibv_reg_mr", -1);
			return -1;
		}
		__queue_append(*q, mr);
	}
	qsort((*q)->mr, (*q)->size, sizeof(struct ibv_mr *), __mr_cmp);
	return 0;
}

int ha_deregmr(struct ha_connector *ctx)
{
	struct mr_queue **q = NULL;
	int res = 0;
	if (ctx->acceptor)
		q = &ctx->acceptor->mr_q;
	else
		q = &((struct ha_connector *)ctx)->mr_q;

	if (!*q) {
		errno = EINVAL;
		return -1;
	}
	for (size_t i = 0; i < (*q)->size; ++i) {
		if (!(*q)->mr[i])
			continue;

		res = ibv_dereg_mr((*q)->mr[i]);
		if (res) {
			err("ibv_dereg_mr", res);
			return res;
		}
		(*q)->mr[i] = NULL;
	}
	free((*q)->mr);
	free(*q);
	*q = NULL;
	return 0;
}

struct ha_acceptor *ha_server(struct ha_cfg *cfg)
{
	int res = 0;
	int flags = 0;
	struct rdma_addrinfo hints, *rai = NULL;

	bzero(&hints, sizeof(hints));
	hints.ai_port_space = RDMA_PS_TCP;
	hints.ai_flags = RAI_PASSIVE;
	char port_s[6] = { 0 };

	snprintf(port_s, sizeof(port_s), "%d", cfg->port);
	int code = rdma_getaddrinfo(cfg->ip, port_s, &hints, &rai);

	if (code) {
		err("rdma_getaddrinfo", code);
		return NULL;
	}
	struct ha_acceptor *ctx =
		(struct ha_acceptor *)calloc(1, sizeof(struct ha_acceptor));
	memcpy(&ctx->cfg, cfg, sizeof(struct ha_cfg));
	ctx->chan = rdma_create_event_channel();
	if (!ctx->chan) {
		err("rdma_create_event_channel", -1);
		res = -1;
		goto failed;
	}
	flags = fcntl(ctx->chan->fd, F_GETFL);
	fcntl(ctx->chan->fd, F_SETFL, flags | O_NONBLOCK);
	res = rdma_create_id(ctx->chan, &ctx->cm_id, NULL, RDMA_PS_TCP);
	if (res) {
		err("rdma_create_id", res);
		goto failed;
	}

	res = rdma_bind_addr(ctx->cm_id, rai->ai_src_addr);
	if (res) {
		err("rdma_bind_addr", res);
		goto failed;
	}

	goto out;

failed:
	if (ctx->chan)
		rdma_destroy_event_channel(ctx->chan);

	if (ctx->cm_id)
		rdma_destroy_id(ctx->cm_id);

	free(ctx);
	ctx = NULL;

out:
	rdma_freeaddrinfo(rai);
	return ctx;
}

int ha_listen(struct ha_acceptor *ctx)
{
	struct ibv_srq_init_attr srq_attr;
	int res = 0;

	assert(ctx->cm_id->verbs);
	assert(ctx->cm_id->pd);
	bzero(&srq_attr, sizeof(srq_attr));
	srq_attr.attr.max_sge = ctx->cfg.max_sge;
	srq_attr.attr.max_wr = ctx->cfg.max_wr;
	ctx->srq = ibv_create_srq(ctx->cm_id->pd, &srq_attr);
	if (!ctx->srq) {
		err("ibv_create_srq", -1);
		res = -1;
		goto failed;
	}

	ctx->cq = ibv_create_cq(ctx->cm_id->verbs, ctx->cfg.max_wr * 2, NULL,
				NULL, 0);
	if (!ctx->cq) {
		err("ibv_create_cq", -1);
		res = -1;
		goto failed;
	}

	res = rdma_listen(ctx->cm_id, SOMAXCONN);
	if (res) {
		err("rdma_listen", res);
		goto failed;
	}

	ctx->conn = calloc(1, sizeof(*ctx->conn));
	ctx->conn->srq = ctx->srq;
	ctx->conn->mr_q = ctx->mr_q;
	ctx->conn->cq = ctx->cq;
	ctx->conn->acceptor = ctx;
	TAILQ_INIT(&ctx->conn->head);
	goto out;

failed:
	if (ctx->srq) {
		ibv_destroy_srq(ctx->srq);
		ctx->srq = NULL;
	}
	if (ctx->cq) {
		ibv_destroy_cq(ctx->cq);
		ctx->cq = NULL;
	}

out:
	return res;
}

int ha_accept(struct ha_acceptor *ctx, struct ha_connector **conn)
{
	int res = 0;
	int flags = 0;
	struct ibv_qp_init_attr qp_attr;
	struct rdma_cm_event *event = NULL;
	struct ha_connector *tmp = NULL;
	struct rdma_cm_id *id = NULL;

	res = rdma_get_cm_event(ctx->chan, &event);
	if (res) {
		if (errno == EAGAIN)
			return 0;
		err("rdma_get_cm_event", res);
		return res;
	}

	switch (event->event) {
	case RDMA_CM_EVENT_CONNECT_REQUEST:
		break;
	case RDMA_CM_EVENT_DISCONNECTED:
		id = event->id;
		rdma_ack_cm_event(event);
		return 0;
	default:
		rdma_ack_cm_event(event);
		err("unexpect event", event->event);
		return 0;
	}

	tmp = (struct ha_connector *)calloc(1, sizeof(*tmp));
	memcpy(&tmp->cfg, &ctx->cfg, sizeof(struct ha_cfg));
	tmp->cm_id = event->id;
	tmp->cm_id->context = tmp;
	rdma_ack_cm_event(event);
	tmp->srq = ctx->srq;
	tmp->cq = ctx->cq;
	tmp->acceptor = ctx;

	bzero(&qp_attr, sizeof(qp_attr));
	qp_attr.qp_context = tmp;
	qp_attr.qp_type = IBV_QPT_RC;
	qp_attr.cap.max_send_wr = ctx->cfg.cap_send_wr;
	qp_attr.cap.max_send_sge = ctx->cfg.max_sge;
	qp_attr.cap.max_recv_sge = ctx->cfg.max_sge;
	qp_attr.cap.max_inline_data = 0;
	qp_attr.recv_cq = tmp->cq;
	qp_attr.srq = tmp->srq;
	qp_attr.send_cq = tmp->cq;
	qp_attr.sq_sig_all = IBV_SEND_SIGNALED;
	errno = 0;
	res = rdma_create_qp(tmp->cm_id, NULL, &qp_attr);
	if (res) {
		err("rdma_create_qp", res);
		goto failed;
	}

	res = rdma_accept(tmp->cm_id, NULL);
	if (res) {
		err("rdma_accept", res);
		goto failed;
	}

	assert(tmp->cm_id->channel->fd == ctx->chan->fd);
	res = __poll_ev(tmp->cm_id->channel->fd, 1000, 1);
	if (res <= 0) {
		err("accept timeout", 0);
		res = 0;
		goto failed;
	}

	res = rdma_get_cm_event(tmp->cm_id->channel, &event);
	if (res) {
		err("rdma_get_cm_event", -1);
		goto failed;
	}
	if (event->event != RDMA_CM_EVENT_ESTABLISHED) {
		err("unexpect event, establish", -1);
		res = -1;
		rdma_ack_cm_event(event);
		goto failed;
	}
	rdma_ack_cm_event(event);

	tmp->connected = true;

	assert(ctx->cm_id->verbs);
	flags = fcntl(ctx->cm_id->verbs->async_fd, F_GETFL);
	fcntl(ctx->cm_id->verbs->async_fd, F_SETFL, flags | O_NONBLOCK);

	tmp->qp_num = tmp->cm_id->qp->qp_num;
	*conn = tmp;
	res = 1;
	goto out;

failed:
	if (tmp->cq)
		ibv_destroy_cq(tmp->cq);
	if (tmp->cm_id) {
		if (tmp->cm_id->qp)
			rdma_destroy_qp(tmp->cm_id);

		rdma_destroy_id(tmp->cm_id);
	}
	free(tmp);

out:
	return res;
}

int ha_client(struct ha_cfg *cfg, struct ha_connector **conn, int timeout)
{
	if (timeout < 10)
		timeout = 10;
	int res = 0;
	int flags = 0;
	struct rdma_cm_event *event = NULL;
	struct rdma_addrinfo hints, *rai = NULL;
	struct ibv_srq_init_attr srq_attr;
	struct ha_connector *tmp = NULL;

	bzero(&hints, sizeof(hints));
	hints.ai_port_space = RDMA_PS_TCP;
	hints.ai_flags = 0;
	char port_s[6] = { 0 };

	snprintf(port_s, sizeof(port_s), "%d", cfg->port);
	res = rdma_getaddrinfo(cfg->ip, port_s, &hints, &rai);

	if (res) {
		err("rdma_getaddrinfo", res);
		goto out;
	}
	tmp = (struct ha_connector *)calloc(1, sizeof(*tmp));
	TAILQ_INIT(&tmp->head);

	tmp->chan = rdma_create_event_channel();
	if (!tmp->chan) {
		res = -1;
		err("rdma_create_event_channel", -1);
		goto gc_chan;
	}
	flags = fcntl(tmp->chan->fd, F_GETFL);
	fcntl(tmp->chan->fd, F_SETFL, flags | O_NONBLOCK);
	res = rdma_create_id(tmp->chan, &tmp->cm_id, tmp, RDMA_PS_TCP);
	if (res) {
		err("rdma_create_id", -1);
		goto gc_chan;
	}

	res = rdma_resolve_addr(tmp->cm_id, NULL, rai->ai_dst_addr, timeout);
	if (res) {
		err("rdma_resolve_addr", res);
		goto gc_id;
	}

	res = __poll_ev(tmp->chan->fd, 1, timeout);
	if (res <= 0) {
		err("resolve addr timeout", -1);
		res = -1;
		goto gc_id;
	}

	res = rdma_get_cm_event(tmp->chan, &event);
	if (res) {
		err("rdma_get_cm_event", res);
		goto gc_id;
	}

	if (event->event != RDMA_CM_EVENT_ADDR_RESOLVED) {
		err("failed to resolve addr", -1);
		res = -1;
		rdma_ack_cm_event(event);
		goto gc_id;
	}

	res = rdma_ack_cm_event(event);
	if (res) {
		err("rdma_ack_cm_event", res);
		goto gc_id;
	}

	event = NULL;
	res = rdma_resolve_route(tmp->cm_id, timeout);
	if (res) {
		err("rdma_resolve_route", res);
		goto gc_id;
	}

	res = __poll_ev(tmp->chan->fd, 1, timeout);
	if (res <= 0) {
		err("resolve route timeout", -1);
		res = -1;
		goto gc_id;
	}

	res = rdma_get_cm_event(tmp->chan, &event);
	if (res) {
		err("rdma_get_cm_event", res);
		goto gc_id;
	}
	if (event->event != RDMA_CM_EVENT_ROUTE_RESOLVED) {
		err("failed to resolve route", -1);
		rdma_ack_cm_event(event);
		res = -1;
		goto gc_id;
	}
	res = rdma_ack_cm_event(event);
	if (res) {
		err("rdma_ack_cm_event", res);
		goto gc_id;
	}

	memcpy(&tmp->cfg, cfg, sizeof(struct ha_cfg));

	assert(tmp->cm_id->pd);
	assert(tmp->cm_id->verbs);
	bzero(&srq_attr, sizeof(srq_attr));
	srq_attr.attr.max_wr = cfg->max_wr;
	srq_attr.attr.max_sge = cfg->max_sge;
	tmp->srq = ibv_create_srq(tmp->cm_id->pd, &srq_attr);
	if (!tmp->srq) {
		err("ibv_create_srq", -1);
		res = -1;
		goto gc_id;
	}
	tmp->cq = ibv_create_cq(tmp->cm_id->verbs, tmp->cfg.max_wr * 2, NULL,
				NULL, 0);
	if (!tmp->cq) {
		err("ibv_create_cq", -1);
		res = -1;
		goto gc_srq;
	}

	assert(tmp->cm_id->verbs);
	flags = fcntl(tmp->cm_id->verbs->async_fd, F_GETFL);
	fcntl(tmp->cm_id->verbs->async_fd, F_SETFL, flags | O_NONBLOCK);
	*conn = tmp;
	goto out;

gc_srq:
	ibv_destroy_srq(tmp->srq);
gc_id:
	rdma_destroy_id(tmp->cm_id);
gc_chan:
	rdma_destroy_event_channel(tmp->chan);
	free(tmp);
out:
	if (rai)
		rdma_freeaddrinfo(rai);

	return res;
}

int ha_connect(struct ha_connector *ctx, int timeout)
{
	int res = 0;
	struct ibv_qp_init_attr qp_attr;
	struct rdma_cm_event *event = NULL;

	bzero(&qp_attr, sizeof(qp_attr));
	qp_attr.qp_context = ctx;
	qp_attr.qp_type = IBV_QPT_RC;
	qp_attr.cap.max_send_wr = ctx->cfg.cap_send_wr;
	qp_attr.cap.max_send_sge = ctx->cfg.max_sge;
	qp_attr.cap.max_recv_sge = ctx->cfg.max_sge;
	qp_attr.cap.max_inline_data = 0;
	qp_attr.recv_cq = ctx->cq;
	qp_attr.srq = ctx->srq;
	qp_attr.sq_sig_all = 0;
	res = rdma_create_qp(ctx->cm_id, ctx->cm_id->pd, &qp_attr);
	if (res) {
		err("rdma_create_qp", res);
		return res;
	}

	res = rdma_connect(ctx->cm_id, NULL);
	if (res) {
		err("rdma_connect", res);
		goto failed;
	}

	res = __poll_ev(ctx->chan->fd, 1, timeout);
	if (res <= 0) {
		err("wait for connect", -1);
		res = -1;
		goto failed;
	}
	res = rdma_get_cm_event(ctx->chan, &event);
	if (res) {
		err("rdma_get_cm_event", res);
		goto failed;
	}
	if (event->event != RDMA_CM_EVENT_ESTABLISHED) {
		errno = EINVAL;
		err(rdma_event_str(event->event), errno);
		res = -1;
		rdma_ack_cm_event(event);
		goto failed;
	}
	res = rdma_ack_cm_event(event);
	if (res) {
		err("rdma_ack_cm_evnet", res);
		goto failed;
	}

	ctx->connected = true;
	ctx->qp_num = ctx->cm_id->qp->qp_num;
	goto out;

failed:
	if (ctx->cm_id->qp) {
		rdma_destroy_qp(ctx->cm_id);
		ctx->cm_id->qp = NULL;
	}
out:
	return res;
}

void ha_destroy_conn(struct ha_connector *ctx)
{
	if (ctx->connected && ctx->cm_id &&
	    ctx->cm_id->qp->state == IBV_QPS_RTS)
		rdma_disconnect(ctx->cm_id);

	if (ctx->cm_id->qp)
		rdma_destroy_qp(ctx->cm_id);

	errno = 0;
	if (ctx->cm_id)
		rdma_destroy_id(ctx->cm_id);

	if (!ctx->acceptor) {
		if (ctx->cq)
			ibv_destroy_cq(ctx->cq);
		if (ctx->srq)
			ibv_destroy_srq(ctx->srq);
		if (ctx->chan)
			rdma_destroy_event_channel(ctx->chan);
	}

	struct ha_wrid *head = NULL;
	struct ha_wrid *next = NULL;

	while (head) {
		next = TAILQ_NEXT(head, next);
		free(head->sge);
		free(head);
		head = next;
	}

	free(ctx);
}

void ha_destroy_acceptor(struct ha_acceptor *a)
{
	struct ha_wrid *head = NULL;
	struct ha_wrid *next = NULL;

	if (a->conn) {
		head = TAILQ_FIRST(&a->conn->head);
		while (head) {
			next = TAILQ_NEXT(head, next);
			free(head->sge);
			free(head);
			head = next;
		}
		free(a->conn);
	}

	if (a->cq)
		ibv_destroy_cq(a->cq);
	if (a->srq)
		ibv_destroy_srq(a->srq);
	if (a->cm_id)
		rdma_destroy_id(a->cm_id);
	if (a->chan)
		rdma_destroy_event_channel(a->chan);
	free(a);
}

static void __ha_wrid_cleaner(struct ha_wrid *id)
{
	struct ha_wrid_que *head = NULL;

	if (id->ctx->acceptor)
		head = &id->ctx->acceptor->conn->head;
	else
		head = &id->ctx->head;

	TAILQ_REMOVE(head, id, next);
	free(id->sge);
	free(id);
}

int ha_post_send(struct ha_connector *ctx, const struct iovec *iov, int n)
{
	struct ibv_sge *sge = NULL;
	struct ha_wrid *id = NULL;
	struct mr_queue *q = NULL;
	struct ha_wrid_que *head = NULL;

	if (ctx->acceptor) {
		q = ctx->acceptor->mr_q;
		head = &ctx->acceptor->conn->head;
	} else {
		q = ctx->mr_q;
		head = &ctx->head;
	}

	if (n > 0)
		sge = (struct ibv_sge *)calloc(n, sizeof(struct ibv_sge));

	for (int i = 0; i < n; ++i) {
		struct ibv_mr *mr =
			__queue_find(q, iov[i].iov_base, iov[i].iov_len, NULL);
		if (!mr) {
			errno = ENOENT;
			err("can't find corresponding memory region", -1);
			free(sge);
			return -1;
		}
		bzero(&sge[i], sizeof(struct ibv_sge));
		sge[i].addr = (uint64_t)iov[i].iov_base;
		sge[i].length = iov[i].iov_len;
		sge[i].lkey = mr->lkey;
	}

	struct ibv_send_wr wr;
	struct ibv_send_wr *bad_wr = NULL;

	bzero(&wr, sizeof(wr));
	wr.opcode = IBV_WR_SEND;
	wr.send_flags = IBV_SEND_SIGNALED;
	wr.sg_list = sge;
	wr.num_sge = n;
	wr.next = NULL;
	id = (struct ha_wrid *)calloc(1, sizeof(struct ha_wrid));
	id->type = SEND;
	id->ctx = ctx;
	id->sge = sge;
	id->n = n;
	id->iov_cb = __ha_wrid_cleaner;
	wr.wr_id = (uint64_t)id;
	int res = ibv_post_send(ctx->cm_id->qp, &wr, &bad_wr);

	if (res) {
		free(sge);
		free(id);
		return res;
	}
	TAILQ_INSERT_TAIL(head, id, next);
	return 0;
}

int ha_post_recv(struct ha_connector *ctx, const struct iovec *iov, int n)
{
	struct ibv_sge *sge =
		(struct ibv_sge *)calloc(n, sizeof(struct ibv_sge));
	struct mr_queue *q = NULL;
	struct ha_wrid_que *head = NULL;

	if (ctx->acceptor) {
		q = ctx->acceptor->mr_q;
		head = &ctx->acceptor->conn->head;
	} else {
		q = ctx->mr_q;
		head = &ctx->head;
	}

	for (int i = 0; i < n; ++i) {
		struct ibv_mr *mr =
			__queue_find(q, iov[i].iov_base, iov[i].iov_len, NULL);
		if (!mr) {
			free(sge);
			errno = ENOENT;
			err("can't find corresponding memory region", -1);
			return -1;
		}
		sge[i].addr = (uint64_t)iov[i].iov_base;
		sge[i].length = iov[i].iov_len;
		sge[i].lkey = mr->lkey;
	}

	struct ibv_recv_wr wr;
	struct ibv_recv_wr *bad_wr = NULL;

	bzero(&wr, sizeof(wr));
	wr.next = NULL;
	wr.num_sge = n;
	wr.sg_list = sge;
	struct ha_wrid *id =
		(struct ha_wrid *)calloc(1, sizeof(struct ha_wrid));
	id->type = RECV;
	id->ctx = ctx;
	id->n = n;
	id->sge = sge;
	id->iov_cb = __ha_wrid_cleaner;
	wr.wr_id = (uint64_t)id;
	int res = ibv_post_srq_recv(ctx->srq, &wr, &bad_wr);

	if (res) {
		free(sge);
		free(id);
		return res;
	}
	TAILQ_INSERT_TAIL(head, id, next);
	return 0;
}

int ha_poll(struct ha_connector *ctx, struct ibv_wc *wc, int nwc)
{
	return ibv_poll_cq(ctx->cq, nwc, wc);
}